from .command import Command


class NewCommand(Command):
    """
    Creates a new Python project at <path>

    new
        { path : The path to create the project at. }
        { --name : Set the resulting package name. }
        { --src : Use the src layout for the project. }
    """

    def handle(self):
        from poetry.layouts import layout
        from poetry.utils._compat import Path
        from poetry.vcs.git import GitConfig

        if self.option('src'):
            layout_ = layout('src')
        else:
            layout_ = layout('standard')

        path = Path.cwd() / Path(self.argument('path'))
        name = self.option('name')
        if not name:
            name = path.name

        if path.exists():
            if list(path.glob('*')):
                # Directory is not empty. Aborting.
                raise RuntimeError(
                    'Destination <fg=yellow>{}</> '
                    'exists and is not empty'.format(
                        path
                    )
                )

        readme_format = 'rst'

        config = GitConfig()
        author = None
        if config.get('user.name'):
            author = config['user.name']
            author_email = config.get('user.email')
            if author_email:
                author += ' <{}>'.format(author_email)

        layout_ = layout_(name, '0.1.0', author=author, readme_format=readme_format)
        layout_.create(path)

        self.line(
            'Created package <info>{}</> in <fg=blue>{}</>'
            .format(name, path.relative_to(Path.cwd()))
        )



from collections import namedtuple


class License(namedtuple('License', 'id name is_osi_approved is_deprecated')):

    CLASSIFIER_SUPPORTED = {
        # Not OSI Approved
        'Aladdin',
        'CC0-1.0',
        'CECILL-B', 'CECILL-C',
        'NPL-1.0', 'NPL-1.1',
        
        # OSI Approved
        'AFPL',
        'AFL-1.1', 'AFL-1.2', 'AFL-2.0', 'AFL-2.1', 'AFL-3.0',
        'Apache-1.1', 'Apache-2.0',
        'APSL-1.1', 'APSL-1.2', 'APSL-2.0',
        'Artistic-1.0', 'Artistic-2.0',
        'AAL',
        'AGPL-3.0', 'AGPL-3.0-only', 'AGPL-3.0-or-later',
        'BSL-1.0',
        'BSD-2-Clause', 'BSD-3-Clause',
        'CDDL-1.0',
        'CECILL-2.1',
        'CPL-1.0',
        'EFL-1.0', 'EFL-2.0',
        'EPL-1.0', 'EPL-2.0',
        'EUPL-1.1', 'EUPL-1.2',
        'GPL-2.0', 'GPL-2.0+', 'GPL-2.0-only', 'GPL-2.0-or-later',
        'GPL-3.0', 'GPL-3.0+', 'GPL-3.0-only', 'GPL-3.0-or-later',
        'LGPL-2.0', 'LGPL-2.0+', 'LGPL-2.0-only', 'LGPL-2.0-or-later',
        'LGPL-3.0', 'LGPL-3.0+', 'LGPL-3.0-only', 'LGPL-3.0-or-later',
        'MIT',
        'MPL-1.0', 'MPL-1.1', 'MPL-1.2',
        'Nokia',
        'W3C',
        'ZPL-1.0', 'ZPL-2.0', 'ZPL-2.1',
    }

    CLASSIFIER_NAMES = {
        # Not OSI Approved
        'AFPL': 'Aladdin Free Public License (AFPL)',
        'CC0-1.0': 'CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'CECILL-B': 'CeCILL-B Free Software License Agreement (CECILL-B)',
        'CECILL-C': 'CeCILL-C Free Software License Agreement (CECILL-C)',
        'NPL-1.0': 'Netscape Public License (NPL)',
        'NPL-1.1': 'Netscape Public License (NPL)',

        # OSI Approved
        'AFL-1.1': 'Academic Free License (AFL)',
        'AFL-1.2': 'Academic Free License (AFL)',
        'AFL-2.0': 'Academic Free License (AFL)',
        'AFL-2.1': 'Academic Free License (AFL)',
        'AFL-3.0': 'Academic Free License (AFL)',
        'Apache-1.1': 'Apache Software License',
        'Apache-2.0': 'Apache Software License',
        'APSL-1.1': 'Apple Public Source License',
        'APSL-1.2': 'Apple Public Source License',
        'APSL-2.0': 'Apple Public Source License',
        'Artistic-1.0': 'Artistic License',
        'Artistic-2.0': 'Artistic License',
        'AAL': 'Attribution Assurance License',
        'AGPL-3.0': 'GNU Affero General Public License v3',
        'AGPL-3.0-only': 'GNU Affero General Public License v3',
        'AGPL-3.0-or-later': 'GNU Affero General Public License v3 or later (AGPLv3+)',
        'BSL-1.0': 'Boost Software License 1.0 (BSL-1.0)',
        'BSD-2-Clause': 'BSD License',
        'BSD-3-Clause': 'BSD License',
        'CDDL-1.0': 'Common Development and Distribution License 1.0 (CDDL-1.0)',
        'CECILL-2.1': 'CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)',
        'CPL-1.0': 'Common Public License',
        'EPL-1.0': 'Eclipse Public License 1.0 (EPL-1.0)',
        'EFL-1.0': 'Eiffel Forum License',
        'EFL-2.0': 'Eiffel Forum License',
        'EUPL-1.1': 'European Union Public Licence 1.1 (EUPL 1.1)',
        'EUPL-1.2': 'European Union Public Licence 1.2 (EUPL 1.2)',
        'GPL-2.0': 'GNU General Public License v2 (GPLv2)',
        'GPL-2.0-only': 'GNU General Public License v2 (GPLv2)',
        'GPL-2.0+': 'GNU General Public License v2 or later (GPLv2+)',
        'GPL-2.0-or-later': 'GNU General Public License v2 or later (GPLv2+)',
        'GPL-3.0': 'GNU General Public License v3 (GPLv3)',
        'GPL-3.0-only': 'GNU General Public License v3 (GPLv3)',
        'GPL-3.0+': 'GNU General Public License v3 or later (GPLv3+)',
        'GPL-3.0-or-later': 'GNU General Public License v3 or later (GPLv3+)',
        'LGPL-2.0': 'GNU Lesser General Public License v2 (LGPLv2)',
        'LGPL-2.0-only': 'GNU Lesser General Public License v2 (LGPLv2)',
        'LGPL-2.0+': 'GNU Lesser General Public License v2 or later (LGPLv2+)',
        'LGPL-2.0-or-later': 'GNU Lesser General Public License v2 or later (LGPLv2+)',
        'LGPL-3.0': 'GNU Lesser General Public License v3 (LGPLv3)',
        'LGPL-3.0-only': 'GNU Lesser General Public License v3 (LGPLv3)',
        'LGPL-3.0+': 'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'LGPL-3.0-or-later': 'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'MPL-1.0': 'Mozilla Public License 1.0 (MPL)',
        'MPL-1.1': 'Mozilla Public License 1.1 (MPL 1.1)',
        'MPL-2.0': 'Mozilla Public License 2.0 (MPL 2.0)',
        'W3C': 'W3C License',
        'ZPL-1.1': 'Zope Public License',
        'ZPL-2.0': 'Zope Public License',
        'ZPL-2.1': 'Zope Public License',
    }

    @property
    def classifier(self):
        parts = ['License']

        if self.is_osi_approved:
            parts.append('OSI Approved')

        name = self.classifier_name
        if name is not None:
            parts.append(name)

        return ' :: '.join(parts)
    
    @property
    def classifier_name(self):
        if self.id not in self.CLASSIFIER_SUPPORTED:
            if self.is_osi_approved:
                return None

            return 'Other/Proprietary License'

        if self.id in self.CLASSIFIER_NAMES:
            return self.CLASSIFIER_NAMES[self.id]

        return self.name



import json
import os

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen


class Updater:

    BASE_URL = 'https://raw.githubusercontent.com/spdx/license-list-data/master/json/'

    def __init__(self, base_url=BASE_URL):
        self._base_url = base_url

    def dump(self, file=None):
        if file is None:
            file = os.path.join(
                os.path.dirname(__file__),
                'data',
                'licenses.json'
            )

        licenses_url = self._base_url + 'licenses.json'

        with open(file, 'w') as f:
            f.write(
                json.dumps(
                    self.get_licenses(licenses_url),
                    indent=2,
                    sort_keys=True
                )
            )

    def get_licenses(self, url):
        licenses = {}
        with urlopen(url) as r:
            data = json.loads(r.read().decode())

        for info in data['licenses']:
            licenses[info['licenseId']] = [
                info['name'],
                info['isOsiApproved'],
                info['isDeprecatedLicenseId']
            ]

        return licenses



from .init import InitCommand
from .venv_command import VenvCommand


class AddCommand(VenvCommand, InitCommand):
    """
    Add a new dependency to <comment>pyproject.toml</>.

    add
        { name* : Packages to add. }
        { --D|dev : Add package as development dependency. }
        { --git= : The url of the Git repository. }
        { --path= : The path to a dependency. }
        { --E|extras=* : Extras to activate for the dependency. }
        { --optional : Add as an optional dependency. }
        { --python= : Python version( for which the dependencies must be installed. }
        { --platform= : Platforms for which the dependencies must be installed. }
        { --allow-prereleases : Accept prereleases. }
        { --dry-run : Outputs the operations but will not execute anything
                     (implicitly enables --verbose). }
    """

    help = """The add command adds required packages to your <comment>pyproject.toml</> and installs them.

If you do not specify a version constraint, poetry will choose a suitable one based on the available package versions.
"""

    _loggers = [
        'poetry.repositories.pypi_repository'
    ]

    def handle(self):
        from poetry.installation import Installer
        from poetry.semver import parse_constraint

        packages = self.argument('name')
        is_dev = self.option('dev')

        if (self.option('git') or self.option('path') or self.option('extras')) and len(packages) > 1:
            raise ValueError(
                'You can only specify one package '
                'when using the --git or --path options'
            )

        if self.option('git') and self.option('path'):
            raise RuntimeError(
                '--git and --path cannot be used at the same time'
            )

        section = 'dependencies'
        if is_dev:
            section = 'dev-dependencies'

        original_content = self.poetry.file.read()
        content = self.poetry.file.read()
        poetry_content = content['tool']['poetry']

        for name in packages:
            for key in poetry_content[section]:
                if key.lower() == name.lower():
                    raise ValueError(
                        'Package {} is already present'.format(name)
                    )

        if self.option('git') or self.option('path'):
            requirements = {
                packages[0]: ''
            }
        else:
            requirements = self._determine_requirements(
                packages,
                allow_prereleases=self.option('allow-prereleases')
            )
            requirements = self._format_requirements(requirements)

            # validate requirements format
            for constraint in requirements.values():
                parse_constraint(constraint)

        for name, constraint in requirements.items():
            constraint = {
                'version': constraint
            }

            if self.option('git'):
                del constraint['version']

                constraint['git'] = self.option('git')
            elif self.option('path'):
                del constraint['version']

                constraint['path'] = self.option('path')

            if self.option('optional'):
                constraint['optional'] = True

            if self.option('allow-prereleases'):
                constraint['allows-prereleases'] = True

            if self.option('extras'):
                extras = []
                for extra in self.option('extras'):
                    if ' ' in extra:
                        extras += [e.strip() for e in extra.split(' ')]
                    else:
                        extras.append(extra)

                constraint['extras'] = self.option('extras')

            if self.option('python'):
                constraint['python'] = self.option('python')

            if self.option('platform'):
                constraint['platform'] = self.option('platform')

            if len(constraint) == 1 and 'version' in constraint:
                constraint = constraint['version']

            poetry_content[section][name] = constraint

        # Write new content
        self.poetry.file.write(content)

        # Cosmetic new line
        self.line('')

        # Update packages
        self.reset_poetry()

        installer = Installer(
            self.output,
            self.venv,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool
        )

        installer.dry_run(self.option('dry-run'))
        installer.update(True)
        installer.whitelist(requirements)

        try:
            status = installer.run()
        except Exception:
            self.poetry.file.write(original_content)

            raise

        if status != 0 or self.option('dry-run'):
            # Revert changes
            if not self.option('dry-run'):
                self.error(
                    '\n'
                    'Addition failed, reverting pyproject.toml '
                    'to its original content.'
                )

            self.poetry.file.write(original_content)

        return status



# -*- coding: utf-8 -*-
import copy
import re

from typing import Union

from poetry.semver import Version
from poetry.semver import parse_constraint
from poetry.spdx import license_by_id
from poetry.spdx import License
from poetry.utils._compat import Path
from poetry.utils.helpers import canonicalize_name

from .constraints.empty_constraint import EmptyConstraint
from .constraints.generic_constraint import GenericConstraint
from .dependency import Dependency
from .directory_dependency import DirectoryDependency
from .file_dependency import FileDependency
from .vcs_dependency import VCSDependency

AUTHOR_REGEX = re.compile('(?u)^(?P<name>[- .,\w\d\'’"()]+)(?: <(?P<email>.+?)>)?$')


class Package(object):

    AVAILABLE_PYTHONS = {
        '2',
        '2.7',
        '3',
        '3.4', '3.5', '3.6', '3.7'
    }

    supported_link_types = {
        'require': {
            'description': 'requires',
            'method': 'requires'
        },
        'provide': {
            'description': 'provides',
            'method': 'provides'
        }
    }

    def __init__(self, name, version, pretty_version=None):
        """
        Creates a new in memory package.
        """
        self._pretty_name = name
        self._name = canonicalize_name(name)

        if not isinstance(version, Version):
            self._version = Version.parse(version)
            self._pretty_version = pretty_version or version
        else:
            self._version = version
            self._pretty_version = pretty_version or self._version.text

        self.description = ''

        self._authors = []

        self.homepage = None
        self.repository_url = None
        self.keywords = []
        self._license = None
        self.readme = None

        self.source_type = ''
        self.source_reference = ''
        self.source_url = ''

        self.requires = []
        self.dev_requires = []
        self.extras = {}
        self.requires_extras = []

        self.category = 'main'
        self.hashes = []
        self.optional = False

        # Requirements for making it mandatory
        self.requirements = {}

        self.build = None
        self.include = []
        self.exclude = []

        self.classifiers = []

        self._python_versions = '*'
        self._python_constraint = parse_constraint('*')
        self._platform = '*'
        self._platform_constraint = EmptyConstraint()

        self.root_dir = None

        self.develop = False

    @property
    def name(self):
        return self._name

    @property
    def pretty_name(self):
        return self._pretty_name

    @property
    def version(self):
        return self._version

    @property
    def pretty_version(self):
        return self._pretty_version

    @property
    def unique_name(self):
        if self.is_root():
            return self._name

        return self.name + '-' + self._version.text

    @property
    def pretty_string(self):
        return self.pretty_name + ' ' + self.pretty_version
    
    @property
    def full_pretty_version(self):
        if self.source_type not in ['hg', 'git']:
            return self._pretty_version

        # if source reference is a sha1 hash -- truncate
        if len(self.source_reference) == 40:
            return '{} {}'.format(self._pretty_version,
                                  self.source_reference[0:7])

        return '{} {}'.format(self._pretty_version, self.source_reference)

    @property
    def authors(self):  # type: () -> list
        return self._authors

    @property
    def author_name(self):  # type: () -> str
        return self._get_author()['name']

    @property
    def author_email(self):  # type: () -> str
        return self._get_author()['email']

    @property
    def all_requires(self):
        return self.requires + self.dev_requires

    def _get_author(self):  # type: () -> dict
        if not self._authors:
            return {
                'name': None,
                'email': None
            }

        m = AUTHOR_REGEX.match(self._authors[0])

        name = m.group('name')
        email = m.group('email')

        return {
            'name': name,
            'email': email
        }

    @property
    def python_versions(self):
        return self._python_versions

    @python_versions.setter
    def python_versions(self, value):
        self._python_versions = value
        self._python_constraint = parse_constraint(value)

    @property
    def python_constraint(self):
        return self._python_constraint

    @property
    def platform(self):  # type: () -> str
        return self._platform

    @platform.setter
    def platform(self, value):  # type: (str) -> None
        self._platform = value
        self._platform_constraint = GenericConstraint.parse(value)

    @property
    def platform_constraint(self):
        return self._platform_constraint

    @property
    def license(self):
        return self._license

    @license.setter
    def license(self, value):
        if value is None:
            self._license = value
        elif isinstance(value, License):
            self._license = value
        else:
            self._license = license_by_id(value)
    
    @property
    def all_classifiers(self):
        classifiers = copy.copy(self.classifiers)

        # Automatically set python classifiers
        if self.python_versions == '*':
            python_constraint = parse_constraint('~2.7 || ^3.4')
        else:
            python_constraint = self.python_constraint

        for version in sorted(self.AVAILABLE_PYTHONS):
            if len(version) == 1:
                constraint = parse_constraint(version + '.*')
            else:
                constraint = Version.parse(version)

            if python_constraint.allows_any(constraint):
                classifiers.append(
                    'Programming Language :: Python :: {}'.format(version)
                )

        # Automatically set license classifiers
        if self.license:
            classifiers.append(self.license.classifier)

        classifiers = set(classifiers)

        return sorted(classifiers)

    def is_prerelease(self):
        return self._version.is_prerelease()

    def is_root(self):
        return False

    def add_dependency(self,
                       name,             # type: str
                       constraint=None,  # type: Union[str, dict, None]
                       category='main'   # type: str
                       ):  # type: (...) -> Dependency
        if constraint is None:
            constraint = '*'

        if isinstance(constraint, dict):
            optional = constraint.get('optional', False)
            python_versions = constraint.get('python')
            platform = constraint.get('platform')
            allows_prereleases = constraint.get('allows-prereleases', False)

            if 'git' in constraint:
                # VCS dependency
                dependency = VCSDependency(
                    name,
                    'git', constraint['git'],
                    branch=constraint.get('branch', None),
                    tag=constraint.get('tag', None),
                    rev=constraint.get('rev', None),
                    optional=optional,
                )
            elif 'file' in constraint:
                file_path = Path(constraint['file'])

                dependency = FileDependency(
                    file_path,
                    category=category,
                    base=self.root_dir
                )
            elif 'path' in constraint:
                path = Path(constraint['path'])

                if self.root_dir:
                    is_file = (self.root_dir / path).is_file()
                else:
                    is_file = path.is_file()

                if is_file:
                    dependency = FileDependency(
                        path,
                        category=category,
                        optional=optional,
                        base=self.root_dir
                    )
                else:
                    dependency = DirectoryDependency(
                        path,
                        category=category,
                        optional=optional,
                        base=self.root_dir,
                        develop=constraint.get('develop', False)
                    )
            else:
                version = constraint['version']

                dependency = Dependency(
                    name, version,
                    optional=optional,
                    category=category,
                    allows_prereleases=allows_prereleases
                )

            if python_versions:
                dependency.python_versions = python_versions

            if platform:
                dependency.platform = platform

            if 'extras' in constraint:
                for extra in constraint['extras']:
                    dependency.extras.append(extra)
        else:
            dependency = Dependency(name, constraint, category=category)

        if category == 'dev':
            self.dev_requires.append(dependency)
        else:
            self.requires.append(dependency)

        return dependency

    def to_dependency(self):
        return Dependency(self.name, self._version)

    def __hash__(self):
        return hash((self._name, self._version))

    def __eq__(self, other):
        if not isinstance(other, Package):
            return NotImplemented

        return self._name == other.name and self._version == other.version

    def __str__(self):
        return self.unique_name

    def __repr__(self):
        return '<Package {}>'.format(self.unique_name)



import os

from ..command import Command


class CacheClearCommand(Command):
    """
    Clears poetry's cache.

    cache:clear
        { cache : The name of the cache to clear. }
        { --all : Clear all caches. }
    """

    def handle(self):
        from cachy import CacheManager
        from poetry.locations import CACHE_DIR

        cache = self.argument('cache')

        parts = cache.split(':')
        cache_dir = os.path.join(CACHE_DIR, 'cache', 'repositories', parts[0])
        cache = CacheManager({
            'default': parts[0],
            'serializer': 'json',
            'stores': {
                parts[0]: {
                    'driver': 'file',
                    'path': cache_dir
                }
            }
        })

        if len(parts) == 1:
            if not self.option('all'):
                raise RuntimeError(
                    'Add the --all option if you want to clear all '
                    '{} caches'.format(parts[0])
                )

            if not os.path.exists(cache_dir):
                self.line('No cache entries for {}'.format(parts[0]))
                return 0

            # Calculate number of entries
            entries_count = 0
            for path, dirs, files in os.walk(cache_dir):
                entries_count += len(files)

            delete = self.confirm(
                '<question>Delete {} entries?</>'.format(entries_count)
            )
            if not delete:
                return 0

            cache.flush()
        elif len(parts) == 2:
            raise RuntimeError(
                'Only specifying the package name is not yet supported. '
                'Add a specific version to clear'
            )
        elif len(parts) == 3:
            package = parts[1]
            version = parts[2]

            if not cache.has('{}:{}'.format(package, version)):
                self.line(
                    'No cache entries for {}:{}'.format(package, version)
                )
                return 0

            delete = self.confirm(
                'Delete cache entry {}:{}'.format(package, version)
            )
            if not delete:
                return 0

            cache.forget('{}:{}'.format(package, version))
        else:
            raise ValueError('Invalid cache key')



import re

from typing import List

from ..command import Command


class DebugResolveCommand(Command):
    """
    Debugs dependency resolution.

    debug:resolve
        { package?* : packages to resolve. }
        { --E|extras=* : Extras to activate for the dependency. }
        { --python= : Python version(s) to use for resolution. }
    """

    _loggers = [
        'poetry.repositories.pypi_repository'
    ]

    def handle(self):
        from poetry.packages import Dependency
        from poetry.packages import ProjectPackage
        from poetry.puzzle import Solver
        from poetry.repositories.repository import Repository
        from poetry.semver import parse_constraint

        packages = self.argument('package')

        if not packages:
            package = self.poetry.package
        else:
            requirements = self._determine_requirements(packages)
            requirements = self._format_requirements(requirements)

            # validate requirements format
            for constraint in requirements.values():
                parse_constraint(constraint)

            dependencies = []
            for name, constraint in requirements.items():
                dep = Dependency(name, constraint)
                extras = []
                for extra in self.option('extras'):
                    if ' ' in extra:
                        extras += [e.strip() for e in extra.split(' ')]
                    else:
                        extras.append(extra)

                for ex in extras:
                    dep.extras.append(ex)

                dependencies.append(dep)

            package = ProjectPackage(
                self.poetry.package.name,
                self.poetry.package.version
            )

            package.python_versions = self.option('python') or self.poetry.package.python_versions
            for dep in dependencies:
                package.requires.append(dep)

        solver = Solver(
            package,
            self.poetry.pool,
            Repository(),
            Repository(),
            self.output
        )

        ops = solver.solve()

        self.line('')
        self.line('Resolution results:')
        self.line('')

        for op in ops:
            package = op.package
            self.line(
                '  - <info>{}</info> (<comment>{}</comment>)'
                .format(
                    package.name, package.version
                )
            )

    def _determine_requirements(self, requires):  # type: (List[str]) -> List[str]
        if not requires:
            return []

        requires = self._parse_name_version_pairs(requires)
        result = []
        for requirement in requires:
            if 'version' not in requirement:
                requirement['version'] = '*'

            result.append(
                '{} {}'.format(requirement['name'], requirement['version'])
            )

        return result

    def _parse_name_version_pairs(self, pairs):  # type: (list) -> list
        result = []

        for i in range(len(pairs)):
            pair = re.sub('^([^=: ]+)[=: ](.*)$', '\\1 \\2', pairs[i].strip())
            pair = pair.strip()

            if ' ' in pair:
                name, version = pair.split(' ', 2)
                result.append({
                    'name': name,
                    'version': version
                })
            else:
                result.append({
                    'name': pair
                })

        return result

    def _format_requirements(self, requirements):  # type: (List[str]) -> dict
        requires = {}
        requirements = self._parse_name_version_pairs(requirements)
        for requirement in requirements:
            requires[requirement['name']] = requirement['version']

        return requires



import operator

from typing import Union

from .exceptions import InvalidVersion
from .legacy_version import LegacyVersion
from .version import Version


OP_EQ = operator.eq
OP_LT = operator.lt
OP_LE = operator.le
OP_GT = operator.gt
OP_GE = operator.ge
OP_NE = operator.ne

_trans_op = {
    '=': OP_EQ,
    '==': OP_EQ,
    '<': OP_LT,
    '<=': OP_LE,
    '>': OP_GT,
    '>=': OP_GE,
    '!=': OP_NE
}


def parse(version,      # type: str
          strict=False  # type: bool
          ):  # type:(...) -> Union[Version, LegacyVersion]
    """
    Parse the given version string and return either a :class:`Version` object
    or a LegacyVersion object depending on if the given version is
    a valid PEP 440 version or a legacy version.

    If strict=True only PEP 440 versions will be accepted.
    """
    try:
        return Version(version)
    except InvalidVersion:
        if strict:
            raise

        return LegacyVersion(version)



from poetry.utils._compat import Path
from poetry.utils.helpers import module_name


class Module:

    def __init__(self, name, directory='.'):
        self._name = module_name(name)
        self._in_src = False

        # It must exist either as a .py file or a directory, but not both
        pkg_dir = Path(directory, self._name)
        py_file = Path(directory, self._name + '.py')
        if pkg_dir.is_dir() and py_file.is_file():
            raise ValueError("Both {} and {} exist".format(pkg_dir, py_file))
        elif pkg_dir.is_dir():
            self._path = pkg_dir
            self._is_package = True
        elif py_file.is_file():
            self._path = py_file
            self._is_package = False
        else:
            # Searching for a src module
            src_pkg_dir = Path(directory, 'src', self._name)
            src_py_file = Path(directory, 'src', self._name + '.py')

            if src_pkg_dir.is_dir() and src_py_file.is_file():
                raise ValueError(
                    "Both {} and {} exist".format(pkg_dir, py_file))
            elif src_pkg_dir.is_dir():
                self._in_src = True
                self._path = src_pkg_dir
                self._is_package = True
            elif src_py_file.is_file():
                self._in_src = True
                self._path = src_py_file
                self._is_package = False
            else:
                raise ValueError("No file/folder found for package {}".format(name))

    @property
    def name(self):  # type: () -> str
        return self._name

    @property
    def path(self):  # type: () -> Path
        return self._path

    @property
    def file(self):  # type: () -> Path
        if self._is_package:
            return self._path / '__init__.py'
        else:
            return self._path

    def is_package(self):  # type: () -> bool
        return self._is_package

    def is_in_src(self):  # type: () -> bool
        return self._in_src



import hashlib
import io

import pkginfo

from pkginfo.distribution import HEADER_ATTRS
from pkginfo.distribution import HEADER_ATTRS_2_0

from poetry.utils._compat import Path

from .dependency import Dependency

# Patching pkginfo to support Metadata version 2.1 (PEP 566)
HEADER_ATTRS.update(
    {
        '2.1': HEADER_ATTRS_2_0 + (
            ('Provides-Extra', 'provides_extra', True),
        )
    }
)


class FileDependency(Dependency):

    def __init__(self,
                 path,             # type: Path
                 category='main',  # type: str
                 optional=False,   # type: bool
                 base=None         # type: Path
                 ):
        self._path = path
        self._base = base
        self._full_path = path

        if self._base and not self._path.is_absolute():
            self._full_path = self._base / self._path

        if not self._full_path.exists():
            raise ValueError('File {} does not exist'.format(self._path))

        if self._full_path.is_dir():
            raise ValueError(
                '{} is a directory, expected a file'.format(self._path)
            )

        if self._path.suffix == '.whl':
            self._meta = pkginfo.Wheel(str(self._full_path))
        else:
            # Assume sdist
            self._meta = pkginfo.SDist(str(self._full_path))

        super(FileDependency, self).__init__(
            self._meta.name,
            self._meta.version,
            category=category,
            optional=optional,
            allows_prereleases=True
        )

    @property
    def path(self):
        return self._path

    @property
    def full_path(self):
        return self._full_path.resolve()

    @property
    def metadata(self):
        return self._meta

    def is_file(self):
        return True

    def hash(self):
        h = hashlib.sha256()
        with self._full_path.open('rb') as fp:
            for content in iter(lambda: fp.read(io.DEFAULT_BUFFER_SIZE), b''):
                h.update(content)

        return h.hexdigest()



from typing import Any

from .incompatibility import Incompatibility
from .term import Term


class Assignment(Term):
    """
    A term in a PartialSolution that tracks some additional metadata.
    """

    def __init__(self, dependency, is_positive,
                 decision_level, index,
                 cause=None):
        super(Assignment, self).__init__(dependency, is_positive)

        self._decision_level = decision_level
        self._index = index
        self._cause = cause

    @property
    def decision_level(self):  # type: () -> int
        return self._decision_level

    @property
    def index(self):  # type: () -> int
        return self._index

    @property
    def cause(self):  # type: () -> Incompatibility
        return self._cause

    @classmethod
    def decision(cls, package, decision_level, index
                 ):  # type: (Any, int, int) -> Assignment
        return cls(package.to_dependency(), True, decision_level, index)

    @classmethod
    def derivation(cls, dependency, is_positive, cause, decision_level, index
                   ):  # type: (Any, bool, Incompatibility, int, int) -> Assignment
        return cls(dependency, is_positive, decision_level, index, cause)

    def is_decision(self):  # type: () -> bool
        return self._cause is None




import os
import pkginfo
import shutil
import time

from cleo import ProgressIndicator
from contextlib import contextmanager
from functools import cmp_to_key
from tempfile import mkdtemp
from typing import List
from typing import Union

from poetry.packages import Dependency
from poetry.packages import DirectoryDependency
from poetry.packages import FileDependency
from poetry.packages import Package
from poetry.packages import VCSDependency
from poetry.packages import dependency_from_pep_508

from poetry.mixology.incompatibility import Incompatibility
from poetry.mixology.incompatibility_cause import DependencyCause
from poetry.mixology.incompatibility_cause import PlatformCause
from poetry.mixology.incompatibility_cause import PythonCause
from poetry.mixology.term import Term

from poetry.repositories import Pool

from poetry.utils._compat import Path
from poetry.utils.helpers import parse_requires
from poetry.utils.toml_file import TomlFile
from poetry.utils.venv import Venv

from poetry.vcs.git import Git

from .dependencies import Dependencies


class Indicator(ProgressIndicator):

    def __init__(self, output):
        super(Indicator, self).__init__(output)

        self.format = '%message% <fg=black;options=bold>(%elapsed:2s%)</>'

    @contextmanager
    def auto(self):
        message = '<info>Resolving dependencies</info>...'

        with super(Indicator, self).auto(message, message):
            yield

    def _formatter_elapsed(self):
        elapsed = time.time() - self.start_time

        return '{:.1f}s'.format(elapsed)


class Provider:

    UNSAFE_PACKAGES = {'setuptools', 'distribute', 'pip'}

    def __init__(self,
                 package,  # type: Package
                 pool,     # type: Pool
                 io
                 ):
        self._package = package
        self._pool = pool
        self._io = io
        self._python_constraint = package.python_constraint
        self._search_for = {}
        self._is_debugging = self._io.is_debug() or self._io.is_very_verbose()

    @property
    def pool(self):  # type: () -> Pool
        return self._pool

    @property
    def name_for_explicit_dependency_source(self):  # type: () -> str
        return 'pyproject.toml'

    @property
    def name_for_locking_dependency_source(self):  # type: () -> str
        return 'pyproject.lock'

    def is_debugging(self):
        return self._is_debugging

    def name_for(self, dependency):  # type: (Dependency) -> str
        """
        Returns the name for the given dependency.
        """
        return dependency.name

    def search_for(self, dependency):  # type: (Dependency) -> List[Package]
        """
        Search for the specifications that match the given dependency.

        The specifications in the returned list will be considered in reverse
        order, so the latest version ought to be last.
        """
        if dependency.is_root:
            return [self._package]

        if dependency in self._search_for:
            return self._search_for[dependency]

        if dependency.is_vcs():
            packages = self.search_for_vcs(dependency)
        elif dependency.is_file():
            packages = self.search_for_file(dependency)
        elif dependency.is_directory():
            packages = self.search_for_directory(dependency)
        else:
            constraint = dependency.constraint

            packages = self._pool.find_packages(
                dependency.name,
                constraint,
                extras=dependency.extras,
                allow_prereleases=dependency.allows_prereleases()
            )

            packages.sort(
                key=lambda p: (
                    not p.is_prerelease() and not dependency.allows_prereleases(),
                    p.version
                ),
                reverse=True
            )

        self._search_for[dependency] = packages

        return self._search_for[dependency]

    def search_for_vcs(self, dependency):  # type: (VCSDependency) -> List[Package]
        """
        Search for the specifications that match the given VCS dependency.

        Basically, we clone the repository in a temporary directory
        and get the information we need by checking out the specified reference.
        """
        if dependency.vcs != 'git':
            raise ValueError(
                'Unsupported VCS dependency {}'.format(dependency.vcs)
            )

        tmp_dir = Path(
            mkdtemp(prefix='pypoetry-git-{}'.format(dependency.name))
        )

        try:
            git = Git()
            git.clone(dependency.source, tmp_dir)
            git.checkout(dependency.reference, tmp_dir)
            revision = git.rev_parse(
                dependency.reference, tmp_dir
            ).strip()

            if dependency.tag or dependency.rev:
                revision = dependency.reference

            pyproject = TomlFile(tmp_dir / 'pyproject.toml')
            pyproject_content = None
            has_poetry = False
            if pyproject.exists():
                pyproject_content = pyproject.read(True)
                has_poetry = (
                    'tool' in pyproject_content
                    and 'poetry' in pyproject_content['tool']
                )

            if pyproject_content and has_poetry:
                # If a pyproject.toml file exists
                # We use it to get the information we need
                info = pyproject_content['tool']['poetry']

                name = info['name']
                version = info['version']
                package = Package(name, version, version)
                package.source_type = dependency.vcs
                package.source_url = dependency.source
                package.source_reference = dependency.reference
                for req_name, req_constraint in info['dependencies'].items():
                    if req_name == 'python':
                        package.python_versions = req_constraint
                        continue

                    package.add_dependency(req_name, req_constraint)
            else:
                # We need to use setup.py here
                # to figure the information we need
                # We need to place ourselves in the proper
                # folder for it to work
                venv = Venv.create(self._io)

                current_dir = os.getcwd()
                os.chdir(tmp_dir.as_posix())

                try:
                    venv.run(
                        'python', 'setup.py', 'egg_info'
                    )

                    egg_info = list(tmp_dir.glob('*.egg-info'))[0]

                    meta = pkginfo.UnpackedSDist(str(egg_info))

                    if meta.requires_dist:
                        reqs = list(meta.requires_dist)
                    else:
                        reqs = []
                        requires = egg_info / 'requires.txt'
                        if requires.exists():
                            with requires.open() as f:
                                reqs = parse_requires(f.read())

                    package = Package(meta.name, meta.version)

                    for req in reqs:
                        package.requires.append(dependency_from_pep_508(req))
                except Exception:
                    raise
                finally:
                    os.chdir(current_dir)

            package.source_type = 'git'
            package.source_url = dependency.source
            package.source_reference = revision
        except Exception:
            raise
        finally:
            shutil.rmtree(tmp_dir.as_posix())

        return [package]

    def search_for_file(self, dependency
                        ):  # type: (FileDependency) -> List[Package]
        package = Package(dependency.name, dependency.pretty_constraint)
        package.source_type = 'file'
        package.source_url = str(dependency.path)

        package.description = dependency.metadata.summary
        for req in dependency.metadata.requires_dist:
            package.requires.append(dependency_from_pep_508(req))

        if dependency.metadata.requires_python:
            package.python_versions = dependency.metadata.requires_python

        if dependency.metadata.platforms:
            package.platform = ' || '.join(dependency.metadata.platforms)

        package.hashes = [dependency.hash()]

        return [package]

    def search_for_directory(self, dependency
                             ):  # type: (DirectoryDependency) -> List[Package]
        package = dependency.package
        if dependency.extras:
            for extra in dependency.extras:
                if extra in package.extras:
                    for dep in package.extras[extra]:
                        dep.activate()

        return [package]

    def incompatibilities_for(self, package):  # type: (Package) -> List[Incompatibility]
        """
        Returns incompatibilities that encapsulate a given package's dependencies,
        or that it can't be safely selected.

        If multiple subsequent versions of this package have the same
        dependencies, this will return incompatibilities that reflect that. It
        won't return incompatibilities that have already been returned by a
        previous call to _incompatibilities_for().
        """
        if package.source_type in ['git', 'file', 'directory']:
            dependencies = package.requires
        elif package.is_root():
            dependencies = package.all_requires
        else:
            dependencies = self._dependencies_for(package)

        if not self._package.python_constraint.allows_any(package.python_constraint):
            return [
                Incompatibility(
                    [Term(package.to_dependency(), True)],
                    PythonCause(package.python_versions)
                )
            ]

        if not self._package.platform_constraint.matches(package.platform_constraint):
            return [
                Incompatibility(
                    [Term(package.to_dependency(), True)],
                    PlatformCause(package.platform)
                )
            ]

        return [
            Incompatibility([
                Term(package.to_dependency(), True),
                Term(dep, False)
            ], DependencyCause())
            for dep in dependencies
        ]

    def dependencies_for(self, package
                         ):  # type: (Package) -> Union[List[Dependency], Dependencies]
        if package.source_type in ['git', 'file', 'directory']:
            # Information should already be set
            return [
                r for r in package.requires
                if not r.is_optional()
                   and r.name not in self.UNSAFE_PACKAGES
            ]
        else:
            return Dependencies(package, self)

    def _dependencies_for(self, package):  # type: (Package) -> List[Dependency]
        complete_package = self._pool.package(
            package.name, package.version.text,
            extras=package.requires_extras
        )

        # Update package with new information
        package.requires = complete_package.requires
        package.description = complete_package.description
        package.python_versions = complete_package.python_versions
        package.platform = complete_package.platform
        package.hashes = complete_package.hashes
        package.extras = complete_package.extras

        return [
            r for r in package.requires
            if not r.is_optional()
            and self._package.python_constraint.allows_any(r.python_constraint)
            and self._package.platform_constraint.matches(package.platform_constraint)
            and r.name not in self.UNSAFE_PACKAGES
        ]

    # UI

    @property
    def output(self):
        return self._io

    def before_resolution(self):
        self._io.write('<info>Resolving dependencies</>')

        if self.is_debugging():
            self._io.new_line()

    def indicate_progress(self):
        if not self.is_debugging():
            self._io.write('.')

    def after_resolution(self):
        self._io.new_line()

    def debug(self, message, depth=0):
        if self.is_debugging():
            debug_info = str(message)
            debug_info = '\n'.join([
                '<comment>{}:</> {}'.format(str(depth).rjust(4), s)
                for s in debug_info.split('\n')
            ]) + '\n'

            self.output.write(debug_info)

    @contextmanager
    def progress(self):
        if not self._io.is_decorated() or self.is_debugging():
            self.output.writeln('Resolving dependencies...')
            yield
        else:
            indicator = Indicator(self._io)

            with indicator.auto():
                yield



from typing import List

from poetry.mixology import resolve_version
from poetry.mixology.failure import SolveFailure
from poetry.packages.constraints.generic_constraint import GenericConstraint

from poetry.semver import parse_constraint

from .exceptions import SolverProblemError

from .operations import Install
from .operations import Uninstall
from .operations import Update
from .operations.operation import Operation

from .provider import Provider


class Solver:

    def __init__(self, package, pool, installed, locked, io):
        self._package = package
        self._pool = pool
        self._installed = installed
        self._locked = locked
        self._io = io

    def solve(self, use_latest=None):  # type: (...) -> List[Operation]
        provider = Provider(self._package, self._pool, self._io)
        locked = {}
        for package in self._locked.packages:
            locked[package.name] = package

        try:
            result = resolve_version(self._package, provider, locked=locked, use_latest=use_latest)
        except SolveFailure as e:
            raise SolverProblemError(e)

        packages = result.packages
        requested = self._package.all_requires

        for package in packages:
            category, optional, python, platform = self._get_tags_for_package(
                package, packages, requested
            )

            package.category = category
            package.optional = optional

            # If requirements are empty, drop them
            requirements = {}
            if python is not None and python != '*':
                requirements['python'] = python

            if platform is not None and platform != '*':
                requirements['platform'] = platform

            package.requirements = requirements

        operations = []
        for package in packages:
            installed = False
            for pkg in self._installed.packages:
                if package.name == pkg.name:
                    installed = True
                    # Checking version
                    if package.version != pkg.version:
                        operations.append(Update(pkg, package))
                    else:
                        operations.append(
                            Install(package).skip('Already installed')
                        )

                    break

            if not installed:
                operations.append(Install(package))

        # Checking for removals
        for pkg in self._locked.packages:
            remove = True
            for package in packages:
                if pkg.name == package.name:
                    remove = False
                    break

            if remove:
                skip = True
                for installed in self._installed.packages:
                    if installed.name == pkg.name:
                        skip = False
                        break

                op = Uninstall(pkg)
                if skip:
                    op.skip('Not currently installed')

                operations.append(op)

        requested_names = [r.name for r in self._package.all_requires]

        return sorted(
            operations,
            key=lambda o: (
                1 if not o.package.name not in requested_names else 0,
                o.package.name
            )
        )

    def _get_graph_for_package(self, package, packages, requested, original=None):
        graph = {
            package.name: {
                'category': 'dev',
                'optional': True,
                'python_version': None,
                'platform': None,
                'dependencies': {},
                'parents': {},
            },
        }

        roots = []
        for dep in requested:
            if dep.name == package.name:
                roots.append(dep)

        origins = []
        for pkg in packages:
            for dep in pkg.all_requires:
                if original and original.name == pkg.name:
                    # Circular dependency
                    continue

                if dep.name == package.name:
                    origins.append((pkg, dep))

        if roots and (not origins or len(roots) > 1):
            # Root dependency
            if len(roots) == 1:
                root = roots[0]
            else:
                root1 = [r for r in roots if r.category == 'main'][0]
                root2 = [r for r in roots if r.category == 'dev'][0]
                if root1.extras == root2.extras or original is None:
                    root = root1
                else:
                    root1_extra_dependencies = []
                    for extra in root1.extras:
                        if extra in package.extras:
                            for dep in package.extras[extra]:
                                root1_extra_dependencies.append(dep.name)

                    root2_extra_dependencies = []
                    for extra in root2.extras:
                        if extra in package.extras:
                            for dep in package.extras[extra]:
                                root2_extra_dependencies.append(dep.name)

                    if (
                            original.name in root1_extra_dependencies
                            and original.name in root2_extra_dependencies
                    ):
                        root = root1
                    elif original.name in root2_extra_dependencies:
                        root = root2
                    else:
                        root = root1

            category = root.category
            optional = root.is_optional()

            python_version = str(root.python_constraint)
            platform = str(root.platform_constraint)

            graph[package.name]['category'] = category
            graph[package.name]['optional'] = optional
            graph[package.name]['python_version'] = python_version
            graph[package.name]['platform'] = platform

            return graph

        for pkg, dep in origins:
            graph[package.name]['dependencies'][pkg.name] = {
                'constraint': dep.pretty_constraint,
                'python_version': dep.python_versions,
                'platform': dep.platform,
            }
            graph[package.name]['parents'].update(
                self._get_graph_for_package(
                    pkg, packages, requested, original=package
                )
            )

        return graph

    def _get_tags_for_package(self, package, packages, requested):
        graph = self._get_graph_for_package(package, packages, requested)[package.name]

        return self._get_tags_from_graph(graph, packages)

    def _get_tags_from_graph(self, graph, packages):
        category = graph['category']
        optional = graph['optional']
        python_version = graph['python_version']
        platform = graph['platform']

        if not graph['parents']:
            # Root dependency
            return category, optional, python_version, platform

        python_versions = []
        platforms = []

        for parent_name, parent_graph in graph['parents'].items():
            dep_python_version = graph['dependencies'][parent_name]['python_version']
            dep_platform = graph['dependencies'][parent_name]['platform']

            for pkg in packages:
                if pkg.name == parent_name:
                    (top_category,
                     top_optional,
                     top_python_version,
                     top_platform) = self._get_tags_from_graph(parent_graph, packages)

                    if category is None or category != 'main':
                        category = top_category

                    optional = optional and top_optional

                    # Take the most restrictive constraints
                    if top_python_version is not None:
                        if dep_python_version is not None:
                            previous = parse_constraint(dep_python_version)
                            current = parse_constraint(top_python_version)

                            if previous.allows_all(current):
                                python_versions.append(top_python_version)
                            else:
                                python_versions.append(dep_python_version)
                        else:
                            python_versions.append(top_python_version)
                    elif dep_python_version is not None:
                        python_versions.append(dep_python_version)

                    if top_platform is not None:
                        if dep_platform is not None:
                            previous = GenericConstraint.parse(dep_platform)
                            current = GenericConstraint.parse(top_platform)

                            if top_platform != '*' and previous.matches(current):
                                platforms.append(top_platform)
                            else:
                                platforms.append(dep_platform)
                        else:
                            platforms.append(top_platform)
                    elif dep_platform is not None:
                        platforms.append(dep_platform)

                    break

        if not python_versions:
            python_version = None
        else:
            # Find the least restrictive constraint
            python_version = python_versions[0]
            previous = parse_constraint(python_version)
            for constraint in python_versions[1:]:
                current = parse_constraint(constraint)

                if python_version == '*':
                    continue
                elif constraint == '*':
                    python_version = constraint
                elif current.allows_all(previous):
                    python_version = constraint

        if not platforms:
            platform = None
        else:
            platform = platforms[0]
            previous = GenericConstraint.parse(platform)
            for constraint in platforms[1:]:
                current = GenericConstraint.parse(constraint)

                if platform == '*':
                    continue
                elif constraint == '*':
                    platform = constraint
                elif current.matches(previous):
                    platform = constraint

        return category, optional, python_version, platform



import os
import shutil
import subprocess
import sys

from email.parser import Parser

from functools import cmp_to_key

from ..command import Command


class SelfUpdateCommand(Command):
    """
    Updates poetry to the latest version.

    self:update
        { version? : The version to update to. }
        { --preview : Install prereleases. }
    """

    def handle(self):
        from poetry.__version__ import __version__
        from poetry.repositories.pypi_repository import PyPiRepository

        version = self.argument('version')
        if not version:
            version = '>=' + __version__

        repo = PyPiRepository(fallback=False)
        packages = repo.find_packages('poetry', version, allow_prereleases=self.option('preview'))
        if not packages:
            self.line('No release found for the specified version')
            return

        packages.sort(
            key=cmp_to_key(
                lambda x, y:
                0 if x.version == y.version
                else int(x.version < y.version or -1)
            )
        )

        release = None
        for package in reversed(packages):
            if package.is_prerelease():
                if self.option('preview'):
                    release = package

                    break

                continue

            release = package

            break

        if release is None:
            self.line('No new release found')
            return

        if release.version == __version__:
            self.line('You are using the latest version')
            return

        try:
            self.update(release)
        except subprocess.CalledProcessError as e:
            self.line('')
            self.output.block([
                '[CalledProcessError]',
                'An error has occured: {}'.format(str(e)),
                e.output
            ], style='error')

            return e.returncode

    def update(self, release):
        from poetry.utils._compat import Path
        from poetry.utils.helpers import temporary_directory

        version = release.version
        self.line('Updating to <info>{}</info>'.format(version))

        prefix = sys.prefix
        base_prefix = getattr(sys, 'base_prefix', None)
        real_prefix = getattr(sys, 'real_prefix', None)

        prefix_poetry = Path(prefix) / 'bin' / 'poetry'
        if prefix_poetry.exists():
            pip = (prefix_poetry.parent / 'pip').resolve()
        elif (
            base_prefix
            and base_prefix != prefix
            and (Path(base_prefix) / 'bin' / 'poetry').exists()
        ):
            pip = Path(base_prefix) / 'bin' / 'pip'
        elif real_prefix:
            pip = Path(real_prefix) / 'bin' / 'pip'
        else:
            pip = Path(prefix) / 'bin' / 'pip'

            if not pip.exists():
                raise RuntimeError('Unable to determine poetry\'s path')

        with temporary_directory(prefix='poetry-update-') as temp_dir:
            temp_dir = Path(temp_dir)
            dist = temp_dir / 'dist'
            self.line('  - Getting dependencies')
            self.process(
                str(pip), 'install', '-U',
                'poetry=={}'.format(release.version),
                '--target', str(dist)
            )

            self.line('  - Vendorizing dependencies')

            poetry_dir = dist / 'poetry'
            vendor_dir = poetry_dir / '_vendor'

            # Everything, except poetry itself, should
            # be put in the _vendor directory
            for file in dist.glob('*'):
                if file.name.startswith('poetry'):
                    continue

                dest = vendor_dir / file.name
                if file.is_dir():
                    shutil.copytree(str(file), str(dest))
                    shutil.rmtree(str(file))
                else:
                    shutil.copy(str(file), str(dest))
                    os.unlink(str(file))

            wheel_data = dist / 'poetry-{}.dist-info'.format(version) / 'WHEEL'
            with wheel_data.open() as f:
                wheel_data = Parser().parsestr(f.read())

            tag = wheel_data['Tag']

            # Repack everything and install
            self.line('  - Updating <info>poetry</info>')

            shutil.make_archive(
                str(temp_dir / 'poetry-{}-{}'.format(version, tag)),
                format='zip',
                root_dir=str(dist)
            )

            os.rename(
                str(temp_dir / 'poetry-{}-{}.zip'.format(version, tag)),
                str(temp_dir / 'poetry-{}-{}.whl'.format(version, tag)),
            )

            self.process(
                str(pip),
                'install',
                '--upgrade',
                '--no-deps',
                str(temp_dir / 'poetry-{}-{}.whl'.format(version, tag))
            )

            self.line('')
            self.line(
                '<info>poetry</> (<comment>{}</>) '
                'successfully installed!'.format(
                    version
                )
            )

    def process(self, *args):
        return subprocess.check_output(list(args), stderr=subprocess.STDOUT)



from .venv_command import VenvCommand


class ScriptCommand(VenvCommand):
    """
    Executes a script defined in <comment>pyproject.toml</comment>. (<error>Deprecated</error>)

    script
        { script-name : The name of the script to execute }
        { args?* : The command and arguments/options to pass to the script. }
    """

    def handle(self):
        self.line('<warning>script is deprecated use run instead.</warning>')
        self.line('')

        script = self.argument('script-name')
        argv = [script] + self.argument('args')

        scripts = self.poetry.local_config.get('scripts')
        if not scripts:
            raise RuntimeError('No scripts defined in pyproject.toml')

        if script not in scripts:
            raise ValueError('Script {} is not defined'.format(script))

        module, callable_ = scripts[script].split(':')

        src_in_sys_path = 'sys.path.append(\'src\'); '\
            if self._module.is_in_src() else ''

        cmd = ['python', '-c']

        cmd += [
            '"import sys; '
            'from importlib import import_module; '
            'sys.argv = {!r}; {}'
            'import_module(\'{}\').{}()"'.format(
               argv, src_in_sys_path, module, callable_
            )
        ]

        self.venv.run(*cmd, shell=True, call=True)

    @property
    def _module(self):
        from ...masonry.utils.module import Module

        poetry = self.poetry
        package = poetry.package
        path = poetry.file.parent
        module = Module(
            package.name, path.as_posix()
        )
        return module

    def merge_application_definition(self, merge_args=True):
        if self._application is None \
                or (self._application_definition_merged
                    and (self._application_definition_merged_with_args or not merge_args)):
            return

        if merge_args:
            current_arguments = self._definition.get_arguments()
            self._definition.set_arguments(self._application.get_definition().get_arguments())
            self._definition.add_arguments(current_arguments)

        self._application_definition_merged = True
        if merge_args:
            self._application_definition_merged_with_args = True



from poetry.semver import parse_constraint
from poetry.semver import VersionUnion

PYTHON_VERSION = [
    '2.7.*',
    '3.0.*', '3.1.*', '3.2.*', '3.3.*', '3.4.*',
    '3.5.*', '3.6.*', '3.7.*', '3.8.*',
]


def format_python_constraint(constraint):
    """
    This helper will help in transforming
    disjunctive constraint into proper constraint.
    """
    if not isinstance(constraint, VersionUnion):
        return str(constraint)

    formatted = []
    accepted = []

    for version in PYTHON_VERSION:
        version_constraint = parse_constraint(version)
        matches = constraint.allows_any(version_constraint)
        if not matches:
            formatted.append('!=' + version)
        else:
            accepted.append(version)

    # Checking lower bound
    low = accepted[0]

    formatted.insert(0, '>=' + '.'.join(low.split('.')[:2]))

    return ', '.join(formatted)



import toml

from poetry.locations import CONFIG_DIR
from poetry.utils._compat import Path

from .uploader import Uploader


class Publisher:
    """
    Registers and publishes packages to remote repositories.
    """

    def __init__(self, poetry, io):
        self._poetry = poetry
        self._package = poetry.package
        self._io = io
        self._uploader = Uploader(poetry, io)

    @property
    def files(self):
        return self._uploader.files

    def publish(self, repository_name, username, password):
        if repository_name:
            self._io.writeln(
                'Publishing <info>{}</info> (<comment>{}</comment>) '
                'to <fg=cyan>{}</>'.format(
                    self._package.pretty_name,
                    self._package.pretty_version,
                    repository_name
                )
            )
        else:
            self._io.writeln(
                'Publishing <info>{}</info> (<comment>{}</comment>) '
                'to <fg=cyan>PyPI</>'.format(
                    self._package.pretty_name,
                    self._package.pretty_version
                )
            )

        if not repository_name:
            url = 'https://upload.pypi.org/legacy/'
            repository_name = 'pypi'
        else:
            # Retrieving config information
            config_file = Path(CONFIG_DIR) / 'config.toml'

            if not config_file.exists():
                raise RuntimeError(
                    'Config file does not exist. '
                    'Unable to get repository information'
                )

            with config_file.open() as f:
                config = toml.loads(f.read())

            if (
                'repositories' not in config
                or repository_name not in config['repositories']
            ):
                raise RuntimeError(
                    'Repository {} is not defined'.format(repository_name)
                )

            url = config['repositories'][repository_name]['url']

        if not (username and password):
            auth_file = Path(CONFIG_DIR) / 'auth.toml'
            if auth_file.exists():
                with auth_file.open() as f:
                    auth_config = toml.loads(f.read())

                if 'http-basic' in auth_config and repository_name in auth_config['http-basic']:
                    config = auth_config['http-basic'][repository_name]

                    username = config.get('username')
                    password = config.get('password')

        # Requesting missing credentials
        if not username:
            username = self._io.ask('Username:')

        if not password:
            password = self._io.ask_hidden('Password:')

        # TODO: handle certificates

        self._uploader.auth(username, password)

        return self._uploader.upload(url)



from .package import Package


class ProjectPackage(Package):

    def is_root(self):
        return True

    def to_dependency(self):
        dependency = super(ProjectPackage, self).to_dependency()

        dependency.is_root = True

        return dependency



import os
import sys

_ROOT = os.path.dirname(os.path.realpath(__file__))
_VENDOR = os.path.join(_ROOT, '_vendor')

# Add vendored dependencies to path.
sys.path.insert(0, _VENDOR)

from .__version__ import __version__  # noqa



import os
import pkginfo

from pkginfo.distribution import HEADER_ATTRS
from pkginfo.distribution import HEADER_ATTRS_2_0

from poetry.io import NullIO
from poetry.utils._compat import Path
from poetry.utils._compat import decode
from poetry.utils.helpers import parse_requires
from poetry.utils.toml_file import TomlFile
from poetry.utils.venv import NullVenv
from poetry.utils.venv import Venv

from .dependency import Dependency

# Patching pkginfo to support Metadata version 2.1 (PEP 566)
HEADER_ATTRS.update(
    {
        '2.1': HEADER_ATTRS_2_0 + (
            ('Provides-Extra', 'provides_extra', True),
        )
    }
)


class DirectoryDependency(Dependency):

    def __init__(self,
                 path,             # type: Path
                 category='main',  # type: str
                 optional=False,   # type: bool
                 base=None,        # type: Path
                 develop=False     # type: bool
                 ):
        from . import dependency_from_pep_508
        from .package import Package

        self._path = path
        self._base = base
        self._full_path = path
        self._develop = develop

        if self._base and not self._path.is_absolute():
            self._full_path = self._base / self._path

        if not self._full_path.exists():
            raise ValueError('Directory {} does not exist'.format(self._path))

        if self._full_path.is_file():
            raise ValueError(
                '{} is a file, expected a directory'.format(self._path)
            )

        # Checking content to dertermine actions
        setup = self._full_path / 'setup.py'
        pyproject = TomlFile(self._full_path / 'pyproject.toml')
        has_poetry = False
        if pyproject.exists():
            pyproject_content = pyproject.read(True)
            has_poetry = (
                    'tool' in pyproject_content
                    and 'poetry' in pyproject_content['tool']
            )

        if not setup.exists() and not has_poetry:
            raise ValueError(
                'Directory {} does not seem to be a Python package'.format(
                    self._full_path
                )
            )

        if has_poetry:
            from poetry.masonry.builders import SdistBuilder
            from poetry.poetry import Poetry

            poetry = Poetry.create(self._full_path)
            builder = SdistBuilder(poetry, NullVenv(), NullIO())

            with setup.open('w') as f:
                f.write(decode(builder.build_setup()))

            package = poetry.package
            self._package = Package(package.pretty_name, package.version)
            self._package.requires += package.requires
            self._package.dev_requires += package.dev_requires
            self._package.python_versions = package.python_versions
            self._package.platform = package.platform
        else:
            # Execute egg_info
            current_dir = os.getcwd()
            os.chdir(str(self._full_path))

            try:
                cwd = base
                venv = Venv.create(NullIO(), cwd=cwd)
                venv.run(
                    'python', 'setup.py', 'egg_info'
                )
            finally:
                os.chdir(current_dir)

            egg_info = list(self._full_path.glob('*.egg-info'))[0]

            meta = pkginfo.UnpackedSDist(str(egg_info))

            if meta.requires_dist:
                reqs = list(meta.requires_dist)
            else:
                reqs = []
                requires = egg_info / 'requires.txt'
                if requires.exists():
                    with requires.open() as f:
                        reqs = parse_requires(f.read())

            package = Package(meta.name, meta.version)
            package.description = meta.summary

            for req in reqs:
                package.requires.append(dependency_from_pep_508(req))

            if meta.requires_python:
                package.python_versions = meta.requires_python

            if meta.platforms:
                platforms = [
                    p
                    for p in meta.platforms
                    if p.lower() != 'unknown'
                ]
                if platforms:
                    package.platform = ' || '.join(platforms)

            self._package = package

        self._package.source_type = 'directory'
        self._package.source_url = str(self._path)

        super(DirectoryDependency, self).__init__(
            self._package.name,
            self._package.version,
            category=category,
            optional=optional,
            allows_prereleases=True
        )

    @property
    def path(self):
        return self._path

    @property
    def full_path(self):
        return self._full_path.resolve()

    @property
    def package(self):
        return self._package

    @property
    def develop(self):
        return self._develop

    def is_directory(self):
        return True



# -*- coding: utf-8 -*-
import re
import subprocess


from poetry.utils._compat import decode


class GitConfig:

    def __init__(self):
        self._config = {}

        try:
            config_list = decode(subprocess.check_output(
                ['git', 'config', '-l'],
                stderr=subprocess.STDOUT
            ))

            m = re.findall('(?ms)^([^=]+)=(.*?)$', config_list)
            if m:
                for group in m:
                    self._config[group[0]] = group[1]
        except subprocess.CalledProcessError:
            pass

    def get(self, key, default=None):
        return self._config.get(key, default)

    def __getitem__(self, item):
        return self._config[item]


class Git:

    def __init__(self, work_dir=None):
        self._config = GitConfig()
        self._work_dir = work_dir

    @property
    def config(self):  # type: () -> GitConfig
        return self._config

    def clone(self, repository, dest):  # type: (...) -> str
        return self.run('clone', repository, str(dest))

    def checkout(self, rev, folder=None):  # type: (...) -> str
        args = []
        if folder is None and self._work_dir:
            folder = self._work_dir

        if folder:
            args += [
                '--git-dir', (folder / '.git').as_posix(),
                '--work-tree', folder.as_posix()
            ]

        args += [
            'checkout', rev
        ]

        return self.run(*args)

    def rev_parse(self, rev, folder=None):  # type: (...) -> str
        args = []
        if folder is None and self._work_dir:
            folder = self._work_dir

        if folder:
            args += [
                '--git-dir', (folder / '.git').as_posix(),
                '--work-tree', folder.as_posix()
            ]

        args += [
            'rev-parse', rev
        ]

        return self.run(*args)

    def get_ignored_files(self, folder=None):  # type: (...) -> list
        args = []
        if folder is None and self._work_dir:
            folder = self._work_dir

        if folder:
            args += [
                '--git-dir', (folder / '.git').as_posix(),
                '--work-tree', folder.as_posix()
            ]

        args += [
            'ls-files', '--others', '-i', '--exclude-standard'
        ]
        output = self.run(*args)

        return output.split('\n')

    def run(self, *args):  # type: (...) -> str
        return decode(subprocess.check_output(
            ['git'] + list(args),
            stderr=subprocess.STDOUT
        ))



import re

MODIFIERS = (
    '[._-]?'
    '((?!post)(?:beta|b|c|pre|RC|alpha|a|patch|pl|p|dev)(?:(?:[.-]?\d+)*)?)?'
    '([+-]?([0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*))?'
)

_COMPLETE_VERSION = 'v?(\d+)(?:\.(\d+))?(?:\.(\d+))?{}(?:\+[^\s]+)?'.format(MODIFIERS)

COMPLETE_VERSION = re.compile('(?i)' + _COMPLETE_VERSION)

CARET_CONSTRAINT = re.compile('(?i)^\^({})$'.format(_COMPLETE_VERSION))
TILDE_CONSTRAINT = re.compile('(?i)^~=?({})$'.format(_COMPLETE_VERSION))
X_CONSTRAINT = re.compile('^(!= ?|==)?v?(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:\.[xX*])+$')
BASIC_CONSTRAINT = re.compile('(?i)^(<>|!=|>=?|<=?|==?)?\s*({})'.format(_COMPLETE_VERSION))



from poetry.utils._compat import Path

from .git import Git


def get_vcs(directory):  # type: (Path) -> Git
    directory = directory.resolve()

    for p in [directory] + list(directory.parents):
        if (p / '.git').is_dir():
            return Git(p)



import re

from .command import Command


class VersionCommand(Command):
    """
    Bumps the version of the project.

    version
        { version=patch }
    """

    help = """\
The version command bumps the version of the project
and writes the new version back to <comment>pyproject.toml</>.

The new version should ideally be a valid semver string or a valid bump rule:
patch, minor, major, prepatch, preminor, premajor, prerelease.
"""

    RESERVED = {
        'major', 'minor', 'patch',
        'premajor', 'preminor', 'prepatch',
        'prerelease'
    }

    def handle(self):
        version = self.argument('version')

        version = self.increment_version(
            self.poetry.package.pretty_version, version
        )

        self.line(
            'Bumping version from <comment>{}</> to <info>{}</>'.format(
                self.poetry.package.pretty_version, version
            )
        )

        content = self.poetry.file.read()
        poetry_content = content['tool']['poetry']
        poetry_content['version'] = version.text

        self.poetry.file.write(content)

    def increment_version(self, version, rule):
        from poetry.semver import Version

        try:
            version = Version.parse(version)
        except ValueError:
            raise ValueError(
                'The project\'s version doesn\'t seem to follow semver'
            )

        if rule in {'major', 'premajor'}:
            new = version.next_major
            if rule == 'premajor':
                new = new.first_prerelease
        elif rule in {'minor', 'preminor'}:
            new = version.next_minor
            if rule == 'preminor':
                new = new.first_prerelease
        elif rule in {'patch', 'prepatch'}:
            new = version.next_patch
            if rule == 'prepatch':
                new = new.first_prerelease
        elif rule == 'prerelease':
            if version.is_prerelease():
                pre = version.prerelease
                new_prerelease = int(pre[1]) + 1
                new = Version.parse(
                    '{}.{}.{}-{}'.format(
                        version.major,
                        version.minor,
                        version.patch,
                        '.'.join([pre[0], str(new_prerelease)])
                    )
                )
            else:
                new = version.next_patch.first_prerelease
        else:
            new = rule

        return new



from .venv_command import VenvCommand


class InstallCommand(VenvCommand):
    """
    Installs the project dependencies.

    install
        { --no-dev : Do not install dev dependencies. }
        { --dry-run : Outputs the operations but will not execute anything
                      (implicitly enables --verbose). }
        { --E|extras=* : Extra sets of dependencies to install. }
        { --develop=* : Install given packages in development mode. }
    """

    help = """The <info>install</info> command reads the <comment>pyproject.toml</> file from
the current directory, processes it, and downloads and installs all the
libraries and dependencies outlined in that file. If the file does not
exist it will look for <comment>pyproject.toml</> and do the same.

<info>poetry install</info>    
"""

    _loggers = [
        'poetry.repositories.pypi_repository'
    ]

    def handle(self):
        from poetry.installation import Installer

        installer = Installer(
            self.output,
            self.venv,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool
        )

        extras = []
        for extra in self.option('extras'):
            if ' ' in extra:
                extras += [e.strip() for e in extra.split(' ')]
            else:
                extras.append(extra)

        installer.extras(extras)
        installer.dev_mode(not self.option('no-dev'))
        installer.develop(self.option('develop'))
        installer.dry_run(self.option('dry-run'))
        installer.verbose(self.option('verbose'))

        return installer.run()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re

from typing import List
from typing import Tuple

from .command import Command
from .venv_command import VenvCommand


class InitCommand(Command):
    """
    Creates a basic <comment>pyproject.toml</> file in the current directory.

    init
        {--name= : Name of the package}
        {--description= : Description of the package}
        {--author= : Author name of the package}
        {--dependency=* : Package to require with an optional version constraint,
                          e.g. requests:^2.10.0 or requests=2.11.1}
        {--dev-dependency=* : Package to require for development with an optional version constraint,
                              e.g. requests:^2.10.0 or requests=2.11.1}
        {--l|license= : License of the package}
    """

    help = """\
The <info>init</info> command creates a basic <comment>pyproject.toml</> file in the current directory.  
"""

    def __init__(self):
        super(InitCommand, self).__init__()

        self._pool = None

    def handle(self):
        from poetry.layouts import layout
        from poetry.utils._compat import Path
        from poetry.vcs.git import GitConfig

        if (Path.cwd() / 'pyproject.toml').exists():
            self.error('A pyproject.toml file already exists.')
            return 1

        vcs_config = GitConfig()

        self.line([
            '',
            'This command will guide you through creating your <info>poetry.toml</> config.',
            ''
        ])

        name = self.option('name')
        if not name:
            name = Path.cwd().name.lower()

            question = self.create_question(
                'Package name [<comment>{}</comment>]: '.format(name),
                default=name
            )
            name = self.ask(question)

        version = '0.1.0'
        question = self.create_question(
            'Version [<comment>{}</comment>]: '.format(version),
            default=version
        )
        version = self.ask(question)

        description = self.option('description') or ''
        question = self.create_question(
            'Description [<comment>{}</comment>]: '.format(description),
            default=description
        )
        description = self.ask(question)

        author = self.option('author')
        if not author and vcs_config and vcs_config.get('user.name'):
            author = vcs_config['user.name']
            author_email = vcs_config.get('user.email')
            if author_email:
                author += ' <{}>'.format(author_email)

        question = self.create_question(
            'Author [<comment>{}</comment>, n to skip]: '.format(author),
            default=author
        )
        question.validator = lambda v: self._validate_author(v, author)
        author = self.ask(question)

        if not author:
            authors = []
        else:
            authors = [author]

        license = self.option('license') or ''

        question = self.create_question(
            'License [<comment>{}</comment>]: '.format(license),
            default=license
        )

        license = self.ask(question)

        question = self.create_question(
            'Compatible Python versions [*]: ',
            default='*'
        )
        python = self.ask(question)

        self.line('')

        requirements = []

        question = 'Would you like to define your dependencies' \
                   ' (require) interactively?'
        if self.confirm(question, True):
            requirements = self._format_requirements(
                self._determine_requirements(self.option('dependency'))
            )

        dev_requirements = []

        question = 'Would you like to define your dev dependencies' \
                   ' (require-dev) interactively'
        if self.confirm(question, True):
            dev_requirements = self._format_requirements(
                self._determine_requirements(self.option('dev-dependency'))
            )

        layout_ = layout('standard')(
            name,
            version,
            description=description,
            author=authors[0] if authors else None,
            license=license,
            python=python,
            dependencies=requirements,
            dev_dependencies=dev_requirements
        )

        content = layout_.generate_poetry_content()
        if self.input.is_interactive():
            self.line('<info>Generated file</info>')
            self.line(['', content, ''])

        if not self.confirm('Do you confirm generation?', True):
            self.line('<error>Command aborted</error>')

            return 1

        with (Path.cwd() / 'pyproject.toml').open('w') as f:
            f.write(content)

    def _determine_requirements(self,
                                requires,  # type: List[str]
                                allow_prereleases=False,  # type: bool
                                ):  # type: (...) -> List[str]
        if not requires:
            requires = []

            package = self.ask('Search for package:')
            while package is not None:
                matches = self._get_pool().search(package)

                if not matches:
                    self.line('<error>Unable to find package</error>')
                    package = False
                else:
                    choices = []

                    for found_package in matches:
                        choices.append(found_package.pretty_name)

                    self.line(
                        'Found <info>{}</info> packages matching <info>{}</info>'
                        .format(
                            len(matches),
                            package
                        )
                    )

                    package = self.choice(
                        '\nEnter package # to add, or the complete package name if it is not listed',
                        choices,
                        attempts=3
                    )

                # no constraint yet, determine the best version automatically
                if package is not False and ' ' not in package:
                    question = self.create_question(
                        'Enter the version constraint to require '
                        '(or leave blank to use the latest version):'
                    )
                    question.attempts = 3
                    question.validator = lambda x: (x or '').strip() or False

                    constraint = self.ask(question)

                    if constraint is False:
                        _, constraint = self._find_best_version_for_package(package)

                        self.line(
                            'Using version <info>{}</info> for <info>{}</info>'
                            .format(constraint, package)
                        )

                    package += ' {}'.format(constraint)

                if package is not False:
                    requires.append(package)

                package = self.ask('\nSearch for a package:')

            return requires

        requires = self._parse_name_version_pairs(requires)
        result = []
        for requirement in requires:
            if 'version' not in requirement:
                # determine the best version automatically
                name, version = self._find_best_version_for_package(
                    requirement['name'],
                    allow_prereleases=allow_prereleases
                )
                requirement['version'] = version
                requirement['name'] = name

                self.line(
                    'Using version <info>{}</> for <info>{}</>'
                    .format(version, name)
                )
            else:
                # check that the specified version/constraint exists
                # before we proceed
                name, _ = self._find_best_version_for_package(
                    requirement['name'], requirement['version'],
                    allow_prereleases=allow_prereleases
                )

                requirement['name'] = name

            result.append(
                '{} {}'.format(requirement['name'], requirement['version'])
            )

        return result

    def _find_best_version_for_package(self,
                                       name,
                                       required_version=None,
                                       allow_prereleases=False
                                       ):  # type: (...) -> Tuple[str, str]
        from poetry.version.version_selector import VersionSelector

        selector = VersionSelector(self._get_pool())
        package = selector.find_best_candidate(
            name, required_version,
            allow_prereleases=allow_prereleases
        )

        if not package:
            # TODO: find similar
            raise ValueError(
                'Could not find a matching version of package {}'.format(name)
            )

        return (
            package.pretty_name,
            selector.find_recommended_require_version(package)
        )

    def _parse_name_version_pairs(self, pairs):  # type: (list) -> list
        result = []

        for i in range(len(pairs)):
            pair = re.sub('^([^=: ]+)[=: ](.*)$', '\\1 \\2', pairs[i].strip())
            pair = pair.strip()

            if ' ' in pair:
                name, version = pair.split(' ', 2)
                result.append({
                    'name': name,
                    'version': version
                })
            else:
                result.append({
                    'name': pair
                })

        return result

    def _format_requirements(self, requirements):  # type: (List[str]) -> dict
        requires = {}
        requirements = self._parse_name_version_pairs(requirements)
        for requirement in requirements:
            requires[requirement['name']] = requirement['version']

        return requires

    def _validate_author(self, author, default):
        from poetry.packages.package import AUTHOR_REGEX

        author = author or default

        if author in ['n', 'no']:
            return

        m = AUTHOR_REGEX.match(author)
        if not m:
            raise ValueError(
                'Invalid author string. Must be in the format: '
                'John Smith <john@example.com>'
            )

        return author

    def _get_pool(self):
        from poetry.repositories import Pool
        from poetry.repositories.pypi_repository import PyPiRepository

        if isinstance(self, VenvCommand):
            return self.poetry.pool

        if self._pool is None:
            self._pool = Pool()
            self._pool.add_repository(PyPiRepository())

        return self._pool



from poetry.toml import dumps
from poetry.toml import loads
from poetry.utils.helpers import module_name


TESTS_DEFAULT = u"""from {package_name} import __version__


def test_version():
    assert __version__ == '{version}'
"""


POETRY_DEFAULT = """\
[tool.poetry]
name = ""
version = ""
description = ""
authors = []

[tool.poetry.dependencies]

[tool.poetry.dev-dependencies]
"""

POETRY_WITH_LICENSE = """\
[tool.poetry]
name = ""
version = ""
description = ""
authors = []
license = ""

[tool.poetry.dependencies]

[tool.poetry.dev-dependencies]
"""


class Layout(object):

    def __init__(self,
                 project,
                 version='0.1.0',
                 description='',
                 readme_format='md',
                 author=None,
                 license=None,
                 python='*',
                 dependencies=None,
                 dev_dependencies=None):
        self._project = project
        self._package_name = module_name(project)
        self._version = version
        self._description = description
        self._readme_format = readme_format
        self._license = license
        self._python = python
        self._dependencies = dependencies or {}
        self._dev_dependencies = dev_dependencies or {'pytest': '^3.5'}

        if not author:
            author = 'Your Name <you@example.com>'

        self._author = author

    def create(self, path, with_tests=True):
        path.mkdir(parents=True, exist_ok=True)

        self._create_default(path)
        self._create_readme(path)

        if with_tests:
            self._create_tests(path)

        self._write_poetry(path)

    def generate_poetry_content(self):
        template = POETRY_DEFAULT
        if self._license:
            template = POETRY_WITH_LICENSE

        content = loads(template)
        poetry_content = content['tool']['poetry']
        poetry_content['name'] = self._project
        poetry_content['version'] = self._version
        poetry_content['description'] = self._description
        poetry_content['authors'].append(self._author)
        if self._license:
            poetry_content['license'] = self._license

        poetry_content['dependencies']['python'] = self._python

        for dep_name, dep_constraint in self._dependencies.items():
            poetry_content['dependencies'][dep_name] = dep_constraint

        for dep_name, dep_constraint in self._dev_dependencies.items():
            poetry_content['dev-dependencies'][dep_name] = dep_constraint

        return dumps(content)

    def _create_default(self, path, src=True):
        raise NotImplementedError()

    def _create_readme(self, path):
        if self._readme_format == 'rst':
            readme_file = path / 'README.rst'
        else:
            readme_file = path / 'README.md'

        readme_file.touch()

    def _create_tests(self, path):
        self._dev_dependencies['pytest'] = '^3.0'

        tests = path / 'tests'
        tests_init = tests / '__init__.py'
        tests_default = tests / 'test_{}.py'.format(self._package_name)

        tests.mkdir()
        tests_init.touch(exist_ok=False)

        with tests_default.open('w') as f:
            f.write(
                TESTS_DEFAULT.format(
                    package_name=self._package_name,
                    version=self._version
                )
            )

    def _write_poetry(self, path):
        content = self.generate_poetry_content()

        poetry = path / 'pyproject.toml'

        with poetry.open('w') as f:
            f.write(content)



import os
import sys

from ..venv_command import VenvCommand


class DebugInfoCommand(VenvCommand):
    """
    Shows debug information.

    debug:info
    """

    def handle(self):
        poetry = self.poetry
        package = poetry.package
        venv = self.venv

        poetry_python_version = '.'.join(str(s) for s in sys.version_info[:3])

        self.output.title('Poetry')
        self.output.listing([
            '<info>Version</info>: <comment>{}</>'.format(poetry.VERSION),
            '<info>Python</info>:  <comment>{}</>'.format(poetry_python_version)
        ])

        self.line('')

        venv_python_version = '.'.join(str(s) for s in venv.version_info[:3])
        self.output.title('Virtualenv')
        self.output.listing([
            '<info>Python</info>:         <comment>{}</>'.format(
                venv_python_version
            ),
            '<info>Implementation</info>: <comment>{}</>'.format(
                venv.python_implementation
            ),
            '<info>Path</info>:           <comment>{}</>'.format(
                venv.venv if venv.is_venv() else 'NA'
            )
        ])

        self.line('')

        self.output.title('System')
        self.output.listing([
            '<info>Platform</info>: <comment>{}</>'.format(sys.platform),
            '<info>OS</info>:       <comment>{}</>'.format(os.name),
        ])

        self.line('')



from typing import Type

from .layout import Layout
from .src import SrcLayout
from .standard import StandardLayout


_LAYOUTS = {
    'src': SrcLayout,
    'standard': StandardLayout,
}


def layout(name):  # type: (str) -> Type[Layout]
    if name not in _LAYOUTS:
        raise ValueError('Invalid layout')

    return _LAYOUTS[name]



# -*- coding: utf-8 -*-
import os
import re
import shutil
import tempfile

from collections import defaultdict
from contextlib import contextmanager

from poetry.utils._compat import Path
from poetry.vcs import get_vcs

from ..metadata import Metadata
from ..utils.module import Module


AUTHOR_REGEX = re.compile('(?u)^(?P<name>[- .,\w\d\'’"()]+) <(?P<email>.+?)>$')


class Builder(object):

    AVAILABLE_PYTHONS = {
        '2',
        '2.7',
        '3',
        '3.4', '3.5', '3.6', '3.7'
    }

    def __init__(self, poetry, venv, io):
        self._poetry = poetry
        self._venv = venv
        self._io = io
        self._package = poetry.package
        self._path = poetry.file.parent
        self._module = Module(
            self._package.name, self._path.as_posix()
        )
        self._meta = Metadata.from_package(self._package)

    def build(self):
        raise NotImplementedError()

    def find_excluded_files(self):  # type: () -> list
        # Checking VCS
        vcs = get_vcs(self._path)
        if not vcs:
            return []

        ignored = vcs.get_ignored_files()
        result = []
        for file in ignored:
            try:
                file = Path(file).absolute().relative_to(self._path)
            except ValueError:
                # Should only happen in tests
                continue

            result.append(file)

        return result

    def find_files_to_add(self, exclude_build=True):  # type: () -> list
        """
        Finds all files to add to the tarball

        TODO: Support explicit include/exclude
        """
        excluded = self.find_excluded_files()
        src = self._module.path
        to_add = []

        if not self._module.is_package():
            if self._module.is_in_src():
                to_add.append(src.relative_to(src.parent.parent))
            else:
                to_add.append(src.relative_to(src.parent))
        else:
            for root, dirs, files in os.walk(src.as_posix()):
                root = Path(root)
                if root.name == '__pycache__':
                    continue

                for file in files:
                    file = root / file
                    file = file.relative_to(self._path)

                    if file in excluded:
                        continue

                    if file.suffix == '.pyc':
                        continue

                    self._io.writeln(
                        ' - Adding: <comment>{}</comment>'.format(str(file)),
                        verbosity=self._io.VERBOSITY_VERY_VERBOSE
                    )
                    to_add.append(file)

        # Include project files
        self._io.writeln(
            ' - Adding: <comment>pyproject.toml</comment>',
            verbosity=self._io.VERBOSITY_VERY_VERBOSE
        )
        to_add.append(Path('pyproject.toml'))

        # If a license file exists, add it
        for license_file in self._path.glob('LICENSE*'):
            self._io.writeln(
                ' - Adding: <comment>{}</comment>'.format(
                    license_file.relative_to(self._path)
                ),
                verbosity=self._io.VERBOSITY_VERY_VERBOSE
            )
            to_add.append(license_file.relative_to(self._path))

        # If a README is specificed we need to include it
        # to avoid errors
        if 'readme' in self._poetry.local_config:
            readme = self._path / self._poetry.local_config['readme']
            if readme.exists():
                self._io.writeln(
                    ' - Adding: <comment>{}</comment>'.format(
                        readme.relative_to(self._path)
                    ),
                    verbosity=self._io.VERBOSITY_VERY_VERBOSE
                )
                to_add.append(readme.relative_to(self._path))

        # If a build script is specified and explicitely required
        # we add it to the list of files
        if self._package.build and not exclude_build:
            to_add.append(Path(self._package.build))

        return sorted(to_add)

    def convert_entry_points(self):  # type: () -> dict
        result = defaultdict(list)

        # Scripts -> Entry points
        for name, ep in self._poetry.local_config.get('scripts', {}).items():
            result['console_scripts'].append('{} = {}'.format(name, ep))

        # Plugins -> entry points
        plugins = self._poetry.local_config.get('plugins', {})
        for groupname, group in plugins.items():
            for name, ep in sorted(group.items()):
                result[groupname].append('{} = {}'.format(name, ep))

        for groupname in result:
            result[groupname] = sorted(result[groupname])

        return dict(result)

    @classmethod
    def convert_author(cls, author):  # type: () -> dict
        m = AUTHOR_REGEX.match(author)

        name = m.group('name')
        email = m.group('email')

        return {
            'name': name,
            'email': email
        }

    @classmethod
    @contextmanager
    def temporary_directory(cls, *args, **kwargs):
        try:
            from tempfile import TemporaryDirectory

            with TemporaryDirectory(*args, **kwargs) as name:
                yield name
        except ImportError:
            name = tempfile.mkdtemp(*args, **kwargs)

            yield name

            shutil.rmtree(name)



import re

from .empty_constraint import EmptyConstraint
from .patterns import BASIC_CONSTRAINT
from .patterns import CARET_CONSTRAINT
from .patterns import TILDE_CONSTRAINT
from .patterns import X_CONSTRAINT
from .version import Version
from .version_constraint import VersionConstraint
from .version_range import VersionRange
from .version_union import VersionUnion


def parse_constraint(constraints):  # type: (str) -> VersionConstraint
    if constraints == '*':
        return VersionRange()

    or_constraints = re.split('\s*\|\|?\s*', constraints.strip())
    or_groups = []
    for constraints in or_constraints:
        and_constraints = re.split(
            '(?<!^)(?<![=>< ,]) *(?<!-)[, ](?!-) *(?!,|$)',
            constraints
        )
        constraint_objects = []

        if len(and_constraints) > 1:
            for constraint in and_constraints:
                constraint_objects.append(parse_single_constraint(constraint))
        else:
            constraint_objects.append(parse_single_constraint(and_constraints[0]))

        if len(constraint_objects) == 1:
            constraint = constraint_objects[0]
        else:
            constraint = constraint_objects[0]
            for next_constraint in constraint_objects[1:]:
                constraint = constraint.intersect(next_constraint)

        or_groups.append(constraint)

    if len(or_groups) == 1:
        return or_groups[0]
    else:
        return VersionUnion.of(*or_groups)


def parse_single_constraint(constraint):  # type: (str) -> VersionConstraint
    m = re.match('(?i)^v?[xX*](\.[xX*])*$', constraint)
    if m:
        return VersionRange()

    # Tilde range
    m = TILDE_CONSTRAINT.match(constraint)
    if m:
        version = Version.parse(m.group(1))

        high = version.stable.next_minor
        if len(m.group(1).split('.')) == 1:
            high = version.stable.next_major

        return VersionRange(version, high, include_min=True)

    # Caret range
    m = CARET_CONSTRAINT.match(constraint)
    if m:
        version = Version.parse(m.group(1))

        return VersionRange(version, version.next_breaking, include_min=True)

    # X Range
    m = X_CONSTRAINT.match(constraint)
    if m:
        op = m.group(1)
        major = int(m.group(2))
        minor = m.group(3)

        if minor is not None:
            version = Version(major, int(minor), 0)

            result = VersionRange(version, version.next_minor, include_min=True)
        else:
            if major == 0:
                result = VersionRange(max=Version(1, 0, 0))
            else:
                version = Version(major, 0, 0)

                result = VersionRange(version, version.next_major, include_min=True)

        if op == '!=':
            result = VersionRange().difference(result)

        return result

    # Basic comparator
    m = BASIC_CONSTRAINT.match(constraint)
    if m:
        op = m.group(1)
        version = m.group(2)

        try:
            version = Version.parse(version)
        except ValueError:
            raise ValueError('Could not parse version constraint: {}'.format(constraint))

        if op == '<':
            return VersionRange(max=version)
        elif op == '<=':
            return VersionRange(max=version, include_max=True)
        elif op == '>':
            return VersionRange(min=version)
        elif op == '>=':
            return VersionRange(min=version, include_min=True)
        elif op == '!=':
            return VersionUnion(
                VersionRange(max=version),
                VersionRange(min=version)
            )
        else:
            return version

    raise ValueError('Could not parse version constraint: {}'.format(constraint))



from typing import Union

from poetry.packages import Package
from poetry.semver import parse_constraint
from poetry.semver import Version


class VersionSelector(object):

    def __init__(self, pool):
        self._pool = pool

    def find_best_candidate(self,
                            package_name,                 # type: str
                            target_package_version=None,  # type:  Union[str, None]
                            allow_prereleases=False       # type: bool
                            ):  # type: (...) -> Union[Package, bool]
        """
        Given a package name and optional version,
        returns the latest Package that matches
        """
        if target_package_version:
            constraint = parse_constraint(target_package_version)
        else:
            constraint = None

        candidates = self._pool.find_packages(
            package_name, constraint,
            allow_prereleases=allow_prereleases
        )

        if not candidates:
            return False

        # Select highest version if we have many
        package = candidates[0]
        for candidate in candidates:
            if candidate.is_prerelease() and not allow_prereleases:
                continue

            # Select highest version of the two
            if package.version < candidate.version:
                package = candidate

        return package

    def find_recommended_require_version(self, package):
        version = package.version

        return self._transform_version(version.text, package.pretty_version)

    def _transform_version(self, version, pretty_version):
        # attempt to transform 2.1.1 to 2.1
        # this allows you to upgrade through minor versions
        try:
            parsed = Version.parse(version)
            parts = [parsed.major, parsed.minor, parsed.patch]
        except ValueError:
            return pretty_version

        # check to see if we have a semver-looking version
        if len(parts) == 3:
            # remove the last parts (the patch version number and any extra)
            if parts[0] != 0:
                del parts[2]

            version = '.'.join(str(p) for p in parts)
            if parsed.is_prerelease():
                version += '-{}'.format('.'.join(str(p) for p in parsed.prerelease))
        else:
            return pretty_version

        return '^{}'.format(version)



from .venv_command import VenvCommand


class LockCommand(VenvCommand):
    """
    Locks the project dependencies.

    lock
    """

    help = """The <info>lock</info> command reads the <comment>pyproject.toml</> file from
the current directory, processes it, and locks the depdencies in the <comment>pyproject.lock</> file.

<info>poetry lock</info>    
"""

    _loggers = [
        'poetry.repositories.pypi_repository'
    ]

    def handle(self):
        from poetry.installation import Installer

        installer = Installer(
            self.output,
            self.venv,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool
        )

        installer.update(True)
        installer.execute_operations(False)

        return installer.run()
